(** * Exponentials in Categories *)
Require Import base ord category pcategory functor products.

Polymorphic Cumulative Class has_exponentials (A : cat) `{has_binary_products A} := mk_has_exponentials
  { Exp : A -> A -> A
  ; happly : forall a b, hom (Prod (Exp a b) b) a
  ; hlam : forall a b c, hom (Prod a b) c -> hom a (Exp c b)
  ; hlam_beta : forall a b c (f : hom (Prod a b) c),
      hfirst (hlam f) ∘ happly c b = f
  ; hlam_eta : forall a b,
      hlam (happly a b) = cid :> hom (Exp a b) (Exp a b)
  ; hlam_nat : forall a b c d (f : hom a b) (g : hom (Prod b c) d),
      hlam (hfirst f ∘ g) = f ∘ hlam g
  }.

Section ExponentialLemmas.
  Variable A : cat.
  Context `{has_exponentials A}.
  Implicit Types (a b c d : A).

  Lemma hlam_betaE a b c d (f : hom a (Prod (Exp b c) c)) (g : hom a d) h :
    f ∘ π₁ = g ∘ hlam h ->
    f ∘ happly b c =  ⟨ g, f ∘ π₂ ⟩ ∘ h.
  Proof.
    move=> E. rewrite <-(hlam_beta h). unfold hfirst. rewrite <-compA.
    rewrite <-(hpair_eta (⟨g, f ∘ π₂⟩ ∘ ⟨π₁ ∘ hlam h, π₂⟩)).
    (* more universe anomalies *)
    csimpl. replace (g ∘ hlam h) with (f ∘ π₁).
    f_equal. symmetry. apply hpair_eta.
  Qed.

  (** Adjunction of homsets *)

  Definition hlam_inv a b c (f : hom a (Exp c b)) : hom (Prod a b) c :=
    hfirst f ∘ happly c b.

  Lemma hlam_retr a b c (f : hom (Prod a b) c) :
    hlam_inv (hlam f) = f.
  Proof. exact: hlam_beta. Qed.

  Lemma hlam_sect a b c (f : hom a (Exp c b)) :
    hlam (hlam_inv f) = f.
  Proof.
    unfold hlam_inv. rewrite-> hlam_nat, hlam_eta. by csimpl.
  Qed.

  Canonical hlam_is_equiv a b c : is_equiv (@hlam A _ _ a b c) :=
    mk_is_equiv (@hlam_inv a b c) (@hlam_retr a b c) (@hlam_sect a b c).
  Canonical hlam_equiv a b c : hom (Prod a b) c <~> hom a (Exp c b) :=
    mk_equiv (@hlam A _ _ a b c) (@hlam_is_equiv a b c).

  (** Naturality *)

  Lemma hlam_nat_second a b b' c (f : hom b b') (g : hom (Prod a b') c) :
    hlam (hsecond f ∘ g) = hlam g ∘ hlam (hsecond f ∘ happly _ _).
  Proof.
    rewrite<-hlam_nat, <-compA, hfirst_hsecond, compA.
      by rewrite-> hlam_beta.
  Qed.

  Lemma hlam_nat_comp a b c c' (f : hom (Prod a b) c) (g : hom c c') :
    hlam (f ∘ g) = hlam f ∘ hlam (happly _ _ ∘ g).
  Proof.
    by rewrite<-hlam_nat, <-compA, hlam_beta.
  Qed.
End ExponentialLemmas.
