Require Import base ord category pcategory functor nat products exponentials yoneda.

(** ** Terminal objects in presheafs *)

Section PresheafTerminal.
  Variable (A : cat).
  Implicit Types (F G : PRESHEAF A).

  Program Definition constant_presheaf (B : Type) : PRESHEAF A :=
    mk_functor (fun=> B) (fun _ _ _ => id) _ _.

  Definition presheaf_terminal : PRESHEAF A := constant_presheaf unit.

  Program Definition presheaf_bang (F : PRESHEAF A) : hom F presheaf_terminal :=
    @mk_nat (OP A) SET F presheaf_terminal (fun _ _ => tt) _.

  Lemma presheaf_terminal_eta (F : PRESHEAF A) (eta : hom F presheaf_terminal) :
    eta = presheaf_bang F.
  Proof.
    apply: nat_eq. fext=> a x /=. by case: (eta a x).
  Qed.
End PresheafTerminal.

Instance presheaf_has_terminal (A : cat) : has_terminal_object (PRESHEAF A) :=
  @mk_has_terminal (PRESHEAF A) (presheaf_terminal A) (@presheaf_bang A) (@presheaf_terminal_eta A).

(** ** Products in presheafs *)

Section PresheafBinaryProducts.
  Variable (A : cat).
  Implicit Types (F G : PRESHEAF A).

  Program Definition PProd F G : PRESHEAF A :=
    mk_functor (fun a => (F a * G a)%type) (fun a b f p => (map F f p.1, map G f p.2)) _ _.
  Next Obligation. fext=> -[x y] /=. by csimpl. Qed.
  Next Obligation. fext=> -[x y] /=. csimpl=>/=. by csimpl. Qed.

  Program Definition nat_pi1 {F G} : hom (PProd F G) F :=
    mk_nat (fun a p => p.1) _.

  Program Definition nat_pi2 {F G} : hom (PProd F G) G :=
    mk_nat (fun a p => p.2) _.

  Program Definition nat_pair {F G H} (eps : hom F G) (eta : hom F H) : hom F (PProd G H) :=
    mk_nat (fun a x => (eps a x, eta a x)) _.
  Next Obligation. fext=> c. csimpl=> /=. by csimpl. Qed.

  Lemma nat_pair1 F G H (eps : hom F G) (eta : hom F H) :
    nat_pair eps eta ∘ nat_pi1 = eps.
  Proof. exact: nat_eq. Qed.

  Lemma nat_pair2 F G H (eps : hom F G) (eta : hom F H) :
    nat_pair eps eta ∘ nat_pi2 = eta.
  Proof. exact: nat_eq. Qed.

  Lemma nat_pair_eta F G H (eta : nat F (PProd G H)) :
    nat_pair (eta ∘ nat_pi1) (eta ∘ nat_pi2) = eta.
  Proof.
    apply: nat_eq. fext=> a x /=. csimpl. by destruct (eta a x).
  Qed.
End PresheafBinaryProducts.

Instance presheaf_has_binary_products (A : cat) : has_binary_products (PRESHEAF A) :=
  mk_has_binary_products (@nat_pair1 A) (@nat_pair2 A) (@nat_pair_eta A).

Instance presheaf_has_products (A : cat) : has_products (PRESHEAF A) :=
  mk_has_products _ _.

(** ** Exponentials in presheafs *)

Section PresheafExponentials.
  Variable A : cat.
  Implicit Types (F G H : PRESHEAF A).

  Local Obligation Tactic := idtac.

  (* Universe anomalies *)
  Definition PExp F G : PRESHEAF A.
    refine (@mk_functor (OP A) SET (fun (a : A) => hom (Prod (y a) G) F) (fun a b (f : A b a) => comp (hfirst (ymap f)))
                        (_ : forall a : OP A, comp (hfirst (ymap cid)) = id) _).
    - abstract (move=> a; fext=> f /=; rewrite-> ymap1; csimpl;
      move: (@hpair_eta_id (PRESHEAF A) (presheaf_has_binary_products A));
      cbn; move=> E; rewrite-> E; exact: nat_idL).
    - abstract (move=> a b c f g; fext=> h; csimpl;
        by rewrite <-ymapC, <-hfirst_comp, <- compA).
  Defined.

  (*Program Definition PExp F G : PRESHEAF A :=
    mk_functor (fun (a : A) => hom (Prod (y a) G) F) (fun a b (f : A b a) g => hfirst (ymap f) ∘ g)
               (_ : forall a : OP A, comp (hfirst (ymap cid)) = id) _.
  Next Obligation.
    move=> F G a. fext=> f /=. rewrite-> ymap1. csimpl.
    move: (@hpair_eta_id (PRESHEAF A) (presheaf_has_binary_products A)).
    cbn. move=> E. rewrite-> E. exact: nat_idL.
  Qed.
  Next Obligation.
    move=> F G a b c f g. fext=> h. csimpl.
      by rewrite <-ymapC, <-hfirst_comp, <- compA.
  Qed.*)

  Program Definition nat_happly F G : hom (Prod (PExp F G) G) F :=
    mk_nat (fun a etap => etap.1 a (cid, etap.2)) _.
  Next Obligation.
    move=> F G a b f. fext=> -[eta x] /=. csimpl=>/=. by csimpl=>/=.
  Qed.

  Program Definition nat_hlam F G H (f : hom (Prod F G) H) : hom F (PExp H G) :=
    mk_nat (fun a x => mk_nat (fun b p => f b (map F p.1 x, p.2)) _) _.
  Next Obligation.
    move=> F G H f a x b c g. fext=> -[/=h y]. csimpl=> /=. by csimpl.
  Qed.
  Next Obligation.
    move=> F G H f a b g. fext=>/=x. csimpl=>/=. apply: nat_eq.
    cbn. fext=> c -[/=h y]. by csimpl=>/=.
  Qed.

  Lemma nat_hlam_beta F G H (f : hom (Prod F G) H) :
    hfirst (nat_hlam f) ∘ nat_happly H G = f.
  Proof.
    apply: nat_eq; fext=> a -[x y] /=. csimpl=>/=. by csimpl.
  Qed.

  Lemma nat_hlam_eta F G :
    nat_hlam (nat_happly F G) = cid.
  Proof.
    apply: nat_eq; fext=> a /= eta. apply: nat_eq; fext=> b [/=f y]. by csimpl=>/=.
  Qed.

  Lemma nat_hlam_nat F G H I (f : hom F G) (g : hom (Prod G H) I) :
    nat_hlam (hfirst f ∘ g) = f ∘ nat_hlam g.
  Proof.
    apply: nat_eq; fext=> a x. apply: nat_eq; fext=> b [/=h y].
      by csimpl=> /=.
  Qed.
End PresheafExponentials.

Instance presheaf_has_exponentials A : @has_exponentials (PRESHEAF A) _ :=
  mk_has_exponentials (@nat_hlam_beta A) (@nat_hlam_eta A) (@nat_hlam_nat A).